from datetime import timedelta
from pathlib import Path
from time import time
from typing import Literal as L
from typing import cast

import click
import humanize
from rich import print
from rich.console import Group
from rich.pretty import Pretty
from rich.table import Table

from torrent_models import TorrentCreate
from torrent_models.compat import get_size
from torrent_models.const import DEFAULT_TORRENT_CREATOR
from torrent_models.create import list_files
from torrent_models.info import InfoDictHybrid, InfoDictHybridCreate, InfoDictV1, InfoDictV2
from torrent_models.torrent import Torrent
from torrent_models.types.common import TorrentVersion
from torrent_models.types.v1 import FileItem, V1PieceLength
from torrent_models.types.v2 import V2PieceLength


@click.group("torrent")
def main() -> None:
    """
    torrent-models CLI
    """


@main.command("make")
@click.option(
    "-p",
    "--path",
    required=True,
    help="Path to a directory or file to create .torrent from",
    type=click.Path(exists=True),
)
@click.option(
    "-t",
    "--tracker",
    required=False,
    default=None,
    multiple=True,
    help="Trackers to add to the torrent. can be used multiple times for multiple trackers. ",
)
@click.option(
    "-s",
    "--piece-size",
    default=512 * (2**10),
    help="Piece size, in bytes",
    show_default=True,
)
@click.option(
    "--comment",
    default=None,
    required=False,
    help="Optional comment field for torrent",
)
@click.option(
    "--creator",
    default=DEFAULT_TORRENT_CREATOR,
    show_default=True,
    required=False,
    help="Optional creator field for torrent",
)
@click.option(
    "-w",
    "--webseed",
    required=False,
    default=None,
    multiple=True,
    help="Add HTTP webseeds as additional sources for torrent. Can be used multiple times. "
    "See https://www.bittorrent.org/beps/bep_0019.html",
)
@click.option(
    "--similar",
    required=False,
    default=None,
    multiple=True,
    help="Add infohash of a similar torrent. "
    "Similar torrents are torrents who have files in common with this torrent, "
    "clients are able to reuse files from the other torrents if they already have them downloaded.",
)
@click.option(
    "--version",
    default="hybrid",
    type=click.Choice(["v1", "v2", "hybrid"]),
    help="What kind of torrent to create, default is hybrid",
)
@click.option("--progress/--no-progress", default=True, help="Enable progress bar (default True)")
@click.option(
    "-o",
    "--output",
    required=False,
    default=None,
    type=click.Path(exists=False),
    help=".torrent file to write to. Otherwise to stdout",
)
def make(
    path: Path,
    tracker: list[str] | tuple[str] | None = None,
    piece_size: V1PieceLength | V2PieceLength = 512 * (2**10),
    comment: str | None = None,
    creator: str = DEFAULT_TORRENT_CREATOR,
    webseed: list[str] | None = None,
    similar: list[str] | None = None,
    version: L["v1", "v2", "hybrid"] = "hybrid",
    progress: bool = True,
    output: Path | None = None,
) -> None:
    path = Path(path)
    files = list_files(path)
    start_time = time()
    created = TorrentCreate(
        trackers=tracker,
        files=files,
        path_root=path,
        comment=comment,
        creator=creator,
        info=InfoDictHybridCreate(piece_length=piece_size, name=path.name),
    )
    generated = created.generate(version=version, progress=progress)
    bencoded = generated.bencode()
    if output:
        with open(output, "wb") as f:
            f.write(bencoded)

        torrent_size = get_size(Path(output))

        end_time = time()
        duration = end_time - start_time
        total_size = generated.info.total_length
        speed = total_size / duration
        click.echo(
            f"Created torrent {output}\n"
            f"Total size: {humanize.naturalsize(total_size, binary=True)}\n"
            f"Torrent size: {humanize.naturalsize(torrent_size, binary=True)}\n"
            f"Duration: {humanize.naturaldelta(timedelta(seconds=duration))}\n"
            f"Speed: {humanize.naturalsize(speed, binary=True)}/s"
        )
    else:
        click.echo(bencoded)


@main.command("print")
@click.argument("torrent", type=click.Path(exists=True))
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="""
              Verbose output, repeat for increasing verbosity\n
              
              -v show files in separate table
              -vv show truncated v1 piece hashes
              -vvv show everything as-is
              """,
)
def pprint(torrent: Path, verbose: int = 0) -> None:
    """
    Print the contents of a torrent file.

    By default, prints only the top-level metadata in a way that should always be
    smaller than one screen.

    Increase verbosity to show more of the torrent.

    Hashes are printed as hexadecimal numbers and split into individual pieces,
    but they are properly encoded in the torrent.
    """
    torrent = Path(torrent)
    t = Torrent.read(torrent)
    # summary stats
    summary = {
        "# Files": humanize.number.intcomma(t.n_files),
        "Total Size": humanize.naturalsize(t.total_size, binary=True),
        "Torrent Size": humanize.naturalsize(get_size(torrent), binary=True),
        "Piece Size": humanize.naturalsize(t.info.piece_length, binary=True),
    }
    v1_infohash = t.v1_infohash
    v2_infohash = t.v2_infohash
    if v1_infohash:
        summary["V1 Infohash"] = v1_infohash.hex()
    if v2_infohash:
        summary["V2 Infohash"] = v2_infohash.hex()
    table = Table(title=str(torrent.name), show_header=False)
    table.add_column("", justify="left", style="magenta bold", no_wrap=True)
    table.add_column("")
    for k, v in summary.items():
        table.add_row(k, v)

    exclude = {}
    context = {"mode": "print", "hash_truncate": True}
    file_table = None
    if verbose <= 1:
        exclude = {"info": {"pieces", "file tree", "file_tree", "files"}, "piece_layers": True}
    elif verbose <= 2:
        exclude = {"info": {"file tree", "file_tree", "files"}, "piece_layers": True}
    else:
        context["hash_truncate"] = False

    # make file table
    if 1 <= verbose <= 2:
        file_table = Table(title="Files")
        file_table.add_column("Path", no_wrap=True)
        file_table.add_column("Size")

        if t.torrent_version == TorrentVersion.v1:
            t.info = cast(InfoDictV1, t.info)
            tfiles = (
                t.info.files
                if t.info.files is not None
                else [FileItem(path=t.info.name, length=t.info.length)]
            )

            files = [
                ("/".join(f.path), humanize.naturalsize(f.length, binary=True), "")
                for f in tfiles
                if f.attr not in (b"p", "p")
            ]
        else:
            t.info = cast(InfoDictV2 | InfoDictHybrid, t.info)
            file_table.add_column("Hash")
            tree = t.flat_files
            assert tree is not None
            files = [
                (
                    str(k),
                    humanize.naturalsize(v["length"], binary=True),
                    v["pieces root"].hex()[0:8],
                )
                for k, v in tree.items()
            ]

        for f in files:
            file_table.add_row(*f)

    dumped = t.model_dump(
        by_alias=True, exclude=exclude, exclude_none=True, context=context  # type: ignore
    )

    if verbose < 1 or verbose > 2:
        group = Group(
            table,
            Pretty(dumped),
        )
    elif verbose <= 2:
        assert file_table is not None
        group = Group(table, Pretty(dumped), file_table)

    print(group)



import bencode_rs
import pytest

from torrent_models import TorrentCreate
from torrent_models.types.common import TorrentVersion

from .conftest import DATA_DIR


def test_create_basic(version):
    """
    Test that we can recreate the basic torrents exactly
    """

    paths = list((DATA_DIR / "basic").rglob("*"))
    create = TorrentCreate(
        paths=paths,
        path_root=DATA_DIR / "basic",
        trackers=["udp://example.com:6969"],
        piece_length=32 * (2**10),
        comment="test",
        created_by="qBittorrent v5.0.4",
        creation_date=1745400513,
        url_list="https://example.com/files",
        info={"source": "source"},
    )
    generated = create.generate(version=version)
    # pdb.set_trace()
    assert generated.torrent_version == TorrentVersion.__members__[version]
    bencoded = generated.bencode()
    with open(DATA_DIR / f"qbt_basic_{version}.torrent", "rb") as f:
        expected = f.read()

    # assert this first for easier error messages
    bdecoded = bencode_rs.bdecode(bencoded)
    expected_decoded = bencode_rs.bdecode(expected)
    assert bdecoded == expected_decoded

    # then test for serialized identity
    assert bencoded == expected


def test_create_libtorrent_static(version):
    """
    Temporary test until we can get libtorrent to stop segfaulting

    ensure torrents we create are identical to libtorrent
    """

    paths = list((DATA_DIR / "basic").rglob("*"))
    create = TorrentCreate(
        paths=paths,
        path_root=DATA_DIR / "basic",
        trackers=["udp://example.com:6969"],
        piece_length=32 * (2**10),
        comment="test",
        created_by="test",
        creation_date=1745400513,
        url_list="https://example.com/files",
    )
    generated = create.generate(version=version)
    assert generated.torrent_version == TorrentVersion.__members__[version]
    bencoded = generated.bencode()
    with open(DATA_DIR / f"lt_basic_{version}.torrent", "rb") as f:
        expected = f.read()

    # assert this first for easier error messages
    bdecoded = bencode_rs.bdecode(bencoded)
    expected_decoded = bencode_rs.bdecode(expected)
    expected_decoded[b"creation date"] = 1745400513
    expected = bencode_rs.bencode(expected_decoded)
    assert bdecoded == expected_decoded

    # then test for serialized identity
    assert bencoded == expected


@pytest.mark.skip(
    reason="libtorrent is segfaulting for some reason... use statically created files"
)
@pytest.mark.libtorrent
def test_create_libtorrent(libtorrent_pair, tmp_path_factory):
    """
    Creating a torrent with libtorrent is the same as creating it with torrent_models
    :param libtorrent_pair:
    :return:
    """
    lt, generated = libtorrent_pair
    assert lt == generated.model_dump_torrent(mode="binary")



from torrent_models.testing.fabricators import (
    default_libtorrent,
    default_tcreate,
    default_torrent,
    make_files,
    make_paths,
)

__all__ = [
    "default_tcreate",
    "default_torrent",
    "default_libtorrent",
    "make_files",
    "make_paths",
]



import hashlib
from math import ceil
from posixpath import join as posixjoin
from typing import Annotated, Any, Self, TypeAlias, cast

import bencode_rs
from annotated_types import Ge, MinLen
from pydantic import BaseModel, Discriminator, Field, Tag, ValidationInfo, model_validator

from torrent_models.base import ConfiguredBase
from torrent_models.types import str_keys
from torrent_models.types.serdes import ByteStr
from torrent_models.types.v1 import FileItem, Pieces, V1PieceLength
from torrent_models.types.v2 import FileTree, FileTreeItem, FileTreeType, V2PieceLength


class InfoDictRoot(ConfiguredBase):
    """Fields shared by v1 and v2 infodicts"""

    name: ByteStr | None = None
    source: ByteStr | None = None

    _total_length: int | None = None

    @property
    def v1_infohash(self) -> bytes | None:
        return None

    @property
    def v2_infohash(self) -> bytes | None:
        return None

    @model_validator(mode="before")
    @classmethod
    def keys_as_strings(cls, data: Any) -> Any:
        """
        bencoded data comes as bytes, if we are trying to create directly from a bytestring dict,
        decode to strings first
        """
        if isinstance(data, dict) and any([isinstance(k, bytes) for k in data]):
            data = str_keys(data)
        return data


class InfoDictV1Base(InfoDictRoot):
    pieces: Pieces | None = None
    length: Annotated[int, Ge(0)] | None = None
    files: Annotated[list[FileItem], MinLen(1)] | None = Field(None)
    piece_length: V1PieceLength | None = Field(alias="piece length")

    @property
    def v1_infohash(self) -> bytes:
        """SHA-1 hash of the infodict"""
        dumped = self.model_dump(exclude_none=True, by_alias=True)
        bencoded = bencode_rs.bencode(dumped)
        return hashlib.sha1(bencoded).digest()

    @property
    def total_length(self) -> int:
        """Total length of all files, in bytes"""
        return self._total_length_v1()

    def _total_length_v1(self) -> int:
        if self._total_length is None:
            if self.files:
                total = 0
                for f in self.files:
                    total += f.length

                self._total_length = total
            else:
                self._total_length = cast(int, self.length)
        return self._total_length

    @model_validator(mode="after")
    def disallowed_fields(self) -> Self:
        """
        We allow extra fields, but not those in v2 infodicts, in order to make them discriminable
        """
        if isinstance(self.__pydantic_extra__, dict):
            assert "file tree" not in self.__pydantic_extra__, "V1 Infodicts can't have file_trees"
        return self

    @model_validator(mode="after")
    def expected_n_pieces(self) -> Self:
        """We have the expected number of pieces given the sizes implied by our file dict"""
        if self.pieces is None or self.piece_length is None:
            return self
        n_pieces = ceil(self.total_length / self.piece_length)
        assert n_pieces == len(self.pieces), (
            f"Expected {n_pieces} pieces for torrent with "
            f"total length {self.total_length} and piece_length"
            f"{self.piece_length}"
            f"Got {len(self.pieces)}"
        )
        return self

    @model_validator(mode="after")
    def padfile_alignment(self, info: ValidationInfo) -> Self:
        """
        If padfiles are present in the files list,
        the sum of a file and its padfile's sizes must be a multiple of the piece size.

        .. note:: V1-only vs hybrid differences

            Some clients do not pad every non-aligned file in v1-only torrents,
            which defeats the purpose of padding, but it happens.
            The default behavior for v1-only is to ignore padfile validation.
            To ensure global padding for v1-only torrents, use pydantic's `strict` validation mode,
            or pass `context = {"padding": "strict"}`.

            Hybrid torrents must have their v1 files list padded,
            and the padding must be globally correct.

        .. note:: Possible Validation Variations

            The behavior of this validator can be changed by passing `padding` to the `context`
            argument of `model_validate` -
            See :class:`~torrent_models.types.validation.ValidationContext`

        """
        if not self.files or len(self.files) == 1 or not self.piece_length:
            return self

        # -- settle switching vars --
        strict = info.config and info.config.get("strict", False)
        hybrid = hasattr(self, "file_tree")

        mode = "default" if not info.context else info.context.get("padding", "default")

        if mode == "default":
            mode = "strict" if hybrid else "ignore"

        # -- do the behavior switch --
        if mode == "ignore" and not strict:
            return self

        if mode == "strict":
            fn = self._validate_padding_strict
        elif mode == "forbid":
            fn = self._validate_padding_forbid
        else:
            raise ValueError(f"unknown padfile validation mode: {mode}")

        for first, second in zip(self.files[:-1], self.files[1:]):
            fn(first, second)

        return self

    def _validate_padding_strict(self, first: FileItem, second: FileItem) -> None:
        # only validate when the first file is not a padfile. if the second file is a padfile,
        # we just validated the pair in the last iteration
        if first.is_padfile:
            return

        # if the first file's length is a multiple of the piece length, no padfile is needed.
        self.piece_length = cast(int, self.piece_length)
        if first.length % self.piece_length == 0:
            return

        # we have a file that needs padding, so second file must be a padfile
        # and the sum must round out
        message = (
            "padding mode: strict - every file that is not a multiple of piece_length "
            "must have a padding file that aligns each file with a piece boundary."
        )

        assert second.is_padfile, message
        assert (first.length + second.length) % self.piece_length == 0, message

    def _validate_padding_forbid(self, first: FileItem, second: FileItem) -> None:
        assert (
            not first.is_padfile and not second.is_padfile
        ), "padding mode: forbid - padfiles are forbidden"


class InfoDictV1(InfoDictV1Base):
    """An infodict from a valid V1 torrent"""

    name: ByteStr
    pieces: Pieces
    piece_length: V1PieceLength = Field(alias="piece length")

    @model_validator(mode="after")
    def length_xor_files(self) -> Self:
        """
        There is also a key length or a key files, but not both or neither.
        If length is present then the download represents a single file,
        otherwise it represents a set of files which go in a directory structure.
        """
        assert bool(self.length) != bool(
            self.files
        ), "V1 Torrents must have a `length` or `files`,  but not both."
        return self


class InfoDictV1Create(InfoDictV1Base):
    """v1 Infodict that may or may not have its pieces hashed yet"""

    pass


class InfoDictV2Base(InfoDictRoot):
    meta_version: int = Field(2, alias="meta version")
    file_tree: FileTreeType | None = Field(None, alias="file tree")
    piece_length: V2PieceLength | None = Field(alias="piece length")

    @model_validator(mode="after")
    def disallowed_fields(self) -> Self:
        """
        We allow extra fields, but not those in v1 infodicts, in order to make them discriminable
        """
        if isinstance(self.__pydantic_extra__, dict):
            assert "pieces" not in self.__pydantic_extra__, "V2 Infodicts can't have pieces"
        return self

    @property
    def v2_infohash(self) -> bytes:
        """SHA-256 hash of the infodict"""
        dumped = self.model_dump(exclude_none=True, by_alias=True)
        bencoded = bencode_rs.bencode(dumped)
        return hashlib.sha256(bencoded).digest()

    @property
    def flat_tree(self) -> dict[str, FileTreeItem]:
        """Flattened file tree! mapping full paths to tree items"""
        if self.file_tree is None:
            return {}
        else:
            return FileTree.flatten_tree(self.file_tree)

    @property
    def total_length(self) -> int:
        """
        Total length of all files, in bytes.
        """
        total_length = 0
        for file in self.flat_tree.values():
            total_length += file["length"]
        return total_length


class InfoDictV2(InfoDictV2Base):
    """An infodict from a valid V2 torrent"""

    name: ByteStr
    piece_length: V2PieceLength = Field(alias="piece length")
    file_tree: FileTreeType = Field(alias="file tree", exclude=False)


class InfoDictV2Create(InfoDictV2Base):
    pass


class InfoDictHybridCreate(InfoDictV1Create, InfoDictV2Create):
    """An infodict of a hybrid torrent that may or may not have its pieces hashed yet"""

    @model_validator(mode="after")
    def disallowed_fields(self) -> Self:
        """hybrids can have any additional fields"""
        return self

    name: ByteStr | None = None
    piece_length: V1PieceLength | V2PieceLength | None = Field(None, alias="piece length")


class InfoDictHybrid(InfoDictV2, InfoDictV1):
    """An infodict of a valid v1/v2 hybrid torrent"""

    piece_length: V2PieceLength = Field(alias="piece length")

    @model_validator(mode="after")
    def disallowed_fields(self) -> Self:
        """hybrids can have any additional fields"""
        return self

    @model_validator(mode="after")
    def expected_n_pieces(self) -> Self:
        """
        We have the expected number of pieces given the sizes implied by our file dict.

        Overrides the v1 to account for expected padding in hybrids
        """
        if self.pieces is None:
            return self
        if self.files is not None:
            n_pieces = ceil(sum([f.length for f in self.files]) / self.piece_length)
        else:
            self.length = cast(int, self.length)
            n_pieces = ceil(self.length / self.piece_length)

        assert n_pieces == len(self.pieces), (
            f"Expected {n_pieces} pieces for hybrid torrent with "
            f"total length {self._total_length_v1()} and piece_length "
            f"{self.piece_length}. "
            f"Got {len(self.pieces)}"
        )
        return self

    @model_validator(mode="after")
    def v1_v2_files_match(self) -> Self:
        """
        From BEP 052:

        > ... the 'pieces' field and 'files' or 'length' in the info dictionary
        > must be generated to describe the same data in the same order.
        > ... Before doing so they must validate that the content
        > (file names, order, piece alignment) is identical.

        file names, sizes, and order must match (ignoring padfiles).
        """
        v2_files = self.flat_tree
        if not self.files:
            v1_files = [FileItem(path=[self.name], length=self.length)]
        else:
            v1_files = [f for f in self.files if not f.is_padfile]

        assert len(v1_files) == len(
            v2_files
        ), "v1 file lists and v2 file trees must have same length"
        for v1_file, v2_item in zip(v1_files, v2_files.items()):
            v2_path, v2_file = v2_item
            v1_posix = posixjoin(*v1_file.path)

            assert v1_posix == v2_path, (
                "v1 file lists and v2 file trees must be in the same order "
                "and have matching path names, excluding v1 padfiles. "
                f"Got:\n"
                f"v1 path: {posixjoin(*v1_file.path)}\n"
                f"v2 path: {v2_path}\n"
            )
            assert v1_file.length == v2_file["length"], (
                "v1 and v2 file lengths must match. Got: \n"
                f"path: {v1_posix}\n"
                f"v1 length: {v1_file.length}\n"
                f"v2 length: {v2_file['length']}"
            )

        return self


def infodict_discriminator(v: Any) -> str | None:
    """
    Discriminator function to use when detecting torrent version,
    and thus which infodict model to validate against.

    Use this instead of standard union discrimination for clearer error messages-
    if there is a validation error in the infodict, since all infodict types will have been tried,
    trivial errors from the two invalid infodict models will also be shown.

    References:
        https://docs.pydantic.dev/latest/concepts/unions/#discriminated-unions-with-callable-discriminator
    """
    # leave the `else` off all these switches to return None if nothing is found.
    # this isn't a validation function, it's just to determine the infodict type
    # if we can't do it here, we do it via validation.
    if isinstance(v, dict):
        v1 = "pieces" in v or b"pieces" in v
        v2 = "file tree" in v or b"file tree" in v
        if v1 and v2:
            return "hybrid"
        elif v1:
            return "v1"
        elif v2:
            return "v2"

    elif isinstance(v, BaseModel):
        if isinstance(v, InfoDictHybrid | InfoDictHybridCreate):
            return "hybrid"
        elif isinstance(v, InfoDictV1Base):
            return "v1"
        elif isinstance(v, InfoDictV2Base):
            return "v2"
    return None


InfodictUnionType: TypeAlias = Annotated[
    (
        Annotated[InfoDictV1, Tag("v1")]
        | Annotated[InfoDictV2, Tag("v2")]
        | Annotated[InfoDictHybrid, Tag("hybrid")]
    ),
    Discriminator(infodict_discriminator),
]



import asyncio
import hashlib
import multiprocessing as mp
from abc import abstractmethod
from collections import deque
from collections.abc import AsyncGenerator
from functools import cached_property
from itertools import count
from math import ceil
from multiprocessing.pool import ApplyResult, AsyncResult
from multiprocessing.pool import Pool as PoolType
from pathlib import Path
from typing import Any, Self, TypeAlias, cast, overload
from typing import Literal as L

from anyio import open_file
from pydantic import BaseModel, Field, model_validator
from tqdm import tqdm

from torrent_models.compat import get_size
from torrent_models.const import BLOCK_SIZE
from torrent_models.types import AbsPath, RelPath, V1PieceLength, V2PieceLength


class Chunk(BaseModel):
    """A single unit of data, usually a 16KiB block, but can be a whole piece e.g. in v1 hashing"""

    path: Path
    """Absolute path"""
    chunk: bytes
    idx: int


class Hash(BaseModel):
    """Hash of a block or piece"""

    type: L["block", "v1_piece", "v2_piece"]
    path: Path
    hash: bytes
    idx: int = Field(
        ...,
        description="""
    The index of the block for ordering.
    
    For v1 hashes, the absolute index of piece across all files.
    For v2 block and piece hashes, index within the given file
    """,
    )


async def iter_blocks(path: Path, read_size: int = BLOCK_SIZE) -> AsyncGenerator[Chunk, None]:
    """Iterate 16KiB blocks"""
    counter = count()
    last_size = read_size
    async with await open_file(path, "rb") as f:
        while last_size == read_size:
            read = await f.read(read_size)
            if len(read) > 0:
                yield Chunk.model_construct(idx=next(counter), path=path, chunk=read)
            last_size = len(read)


class HasherBase(BaseModel):
    paths: list[RelPath]
    """
    Relative paths beneath the path base to hash.
    
    Paths should already be sorted in the order they are to appear in the torrent
    """
    path_root: AbsPath
    """Directory containing paths to hash"""
    piece_length: V1PieceLength | V2PieceLength
    n_processes: int = Field(default_factory=mp.cpu_count)
    progress: bool = False
    """Show progress"""
    read_size: int | None = None
    """
    How much of a file should be read in a single read call.
    
    If None, set to the piece_length
    """
    memory_limit: int | None = None
    """
    Rough cap on outstanding memory usage (in bytes) - pauses reading more data until
    the number of outstanding chunks to process are smaller than this size
    """

    @staticmethod
    def _hash_v1(chunk: Chunk, path_root: Path) -> Hash:
        return Hash.model_construct(
            hash=hashlib.sha1(chunk.chunk).digest(),
            type="v1_piece",
            path=chunk.path.relative_to(path_root),
            idx=chunk.idx,
        )

    @staticmethod
    def _hash_v2(chunk: Chunk, path_root: Path) -> Hash:
        return Hash.model_construct(
            hash=hashlib.sha256(chunk.chunk).digest(),
            type="block",
            path=chunk.path.relative_to(path_root),
            idx=chunk.idx,
        )

    @abstractmethod
    def update(self, chunk: Chunk, pool: PoolType) -> list[AsyncResult]:
        """
        Update hasher with a new chunk of data, returning a list of AsyncResults to fetch hashes
        """
        pass

    @model_validator(mode="after")
    def read_size_defaults_piece_size(self) -> Self:
        if not self.read_size:
            self.read_size = self.piece_length
        return self

    def complete(self, hashes: list[Hash]) -> list[Hash]:
        """After hashing, do any postprocessing to yield the desired output"""
        return hashes

    def _after_read(self, pool: PoolType) -> list[AsyncResult]:
        """Optional step after reading completes"""
        return []

    @cached_property
    def file_sizes(self) -> list[tuple[Path, int]]:
        return [(path, get_size(self.path_root / path)) for path in self.paths]

    @cached_property
    def total_chunks(self) -> int:
        """Total read_size chunks in all files"""
        total_chunks = 0
        self.read_size = cast(int, self.read_size)
        for _, size in self.file_sizes:
            total_chunks += ceil(size / self.read_size)
        return total_chunks

    @cached_property
    def total_size(self) -> int:
        return sum([fs[1] for fs in self.file_sizes])

    @cached_property
    def total_hashes(self) -> int:
        """Total hashes that need to be computed"""
        return self.total_chunks

    def _v1_total_hashes(self) -> int:
        return ceil(self.total_size / self.piece_length)

    def _v2_total_hashes(self) -> int:
        """Total hashes that need to be computed"""
        total_blocks = 0
        for _, size in self.file_sizes:
            total_blocks += ceil(size / BLOCK_SIZE)
        return total_blocks

    def _v1_total_hashes_hybrid(self) -> int:
        total_pieces = 0
        for _, size in self.file_sizes:
            total_pieces += ceil(size / self.piece_length)
        return total_pieces

    @cached_property
    def max_outstanding_results(self) -> int | None:
        """Total number of async result objects that can be outstanding, to limit memory usage"""
        if self.memory_limit is None:
            return None
        else:
            self.read_size = cast(int, self.read_size)
            return self.memory_limit // self.read_size

    async def process_async(self) -> list[Hash]:
        hashes = await self.hash()
        return self.complete(hashes)

    def process(self) -> list[Hash]:
        try:
            return asyncio.run(self.process_async())
        except RuntimeError as e:
            if "from a running event loop" in str(e):
                import nest_asyncio  # type: ignore

                nest_asyncio.apply()
                loop = asyncio.get_event_loop()
                return loop.run_until_complete(self.process_async())
            else:
                raise e

    async def hash(self) -> list[Hash]:
        """
        Hash all files
        """
        with mp.Pool(self.n_processes) as pool:
            pbars = self._pbars()

            hashes = []
            results: deque[ApplyResult] = deque()
            # if "1_1073741824_equal0" in str(self.path_root):
            #     breakpoint()
            try:
                for path in self.paths:

                    pbars.file.set_description(str(path))
                    self.read_size = cast(int, self.read_size)
                    async for chunk in iter_blocks(self.path_root / path, read_size=self.read_size):
                        pbars.read.update()
                        res = self.update(chunk, pool)
                        results.extend(res)
                        results, hash = self._step_results(results)
                        if hash is not None:
                            hashes.append(hash)
                            pbars.hash.update()

                        if self.max_outstanding_results:
                            while len(results) > self.max_outstanding_results:
                                results, hash = self._step_results(results, block=True)
                                hashes.append(hash)
                                pbars.hash.update()

                    pbars.file.update()

                results.extend(self._after_read(pool))
                while len(results) > 0:
                    results, hash = self._step_results(results, block=True)
                    hashes.append(hash)
                    pbars.hash.update()

            finally:
                pbars.close()

        return hashes

    @overload
    def _step_results(self, results: deque, block: L[True]) -> tuple[deque, Hash]: ...

    @overload
    def _step_results(self, results: deque, block: L[False]) -> tuple[deque, Hash | None]: ...

    @overload
    def _step_results(self, results: deque) -> tuple[deque, Hash | None]: ...

    def _step_results(self, results: deque, block: bool = False) -> tuple[deque, Hash | None]:
        """Step the outstanding results, yielding a single hash"""
        if len(results) == 0:
            return results, None

        res = results.popleft()
        if block:
            return results, res.get()
        else:
            try:
                return results, res.get(timeout=0)
            except mp.TimeoutError:
                # she not done yet
                results.appendleft(res)
                return results, None

    def _pbars(self) -> "_PBars":
        return _PBars(
            dummy=not self.progress,
            file_total=len(self.paths),
            read_total=self.total_chunks,
            hash_total=self.total_hashes,
        )


class DummyPbar:
    """pbar that does nothing so we i don't get fined by mypy"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def update(self, n: int = 1) -> None:
        pass

    def close(self) -> None:
        pass

    def set_description(self, *args: Any, **kwargs: Any) -> None:
        pass


PbarLike: TypeAlias = DummyPbar | tqdm


class _PBars:
    """
    Wrapper around multiple pbars, including dummy pbars for when progress is disabled
    """

    def __init__(
        self,
        file_total: int | None = None,
        read_total: int | None = None,
        hash_total: int | None = None,
        dummy: bool = False,
    ):
        self.file: PbarLike
        self.read: PbarLike
        self.hash: PbarLike
        if dummy:
            self.file = DummyPbar()
            self.read = DummyPbar()
            self.hash = DummyPbar()
        else:
            self.file = tqdm(total=file_total, desc="File", position=0)
            self.read = tqdm(total=read_total, desc="Reading Chunk", position=1)
            self.hash = tqdm(total=hash_total, desc="Hashing Chunk", position=2)

    def close(self) -> None:
        self.file.close()
        self.read.close()
        self.hash.close()



import shutil
from pathlib import Path
from random import randbytes

import pytest
from torf import Torrent as TorfTorrent

from torrent_models import Torrent, TorrentCreate, TorrentVersion, size

DATA_DIR = (Path(__file__).parent / "data").resolve()

MULTI_HYBRID = DATA_DIR / "qbt_directory_hybrid.torrent"
MULTI_V1 = DATA_DIR / "qbt_directory_v1.torrent"
MULTI_V2 = DATA_DIR / "qbt_directory_v2.torrent"
GIANT_HYBRID = DATA_DIR / "qbt_giant_hybrid.torrent"
GIANT_V1 = DATA_DIR / "qbt_giant_v1.torrent"
GIANT_V2 = DATA_DIR / "qbt_giant_v2.torrent"

MULTI_FILE_HYBRID = Path(__file__).joinpath("../data/qbt_directory_hybrid.torrent").resolve()
GIANT_TORRENT = Path(__file__).joinpath("../data/qbt_giant_v1.torrent").resolve()


@pytest.fixture(
    params=[
        pytest.param(MULTI_HYBRID, id="multi-hybrid"),
        pytest.param(MULTI_V1, id="multi-v1"),
        pytest.param(MULTI_V2, id="multi-v2"),
        pytest.param(GIANT_HYBRID, id="giant-hybrid"),
        pytest.param(GIANT_V1, id="giant-v1"),
        pytest.param(GIANT_V2, id="giant-v2"),
    ]
)
def torrent_file(request) -> Path:
    if not request.param.exists():
        pytest.skip()
    return request.param


@pytest.fixture(
    params=[
        pytest.param(TorrentVersion.v1, id="v1"),
        pytest.param(TorrentVersion.v2, id="v2"),
        pytest.param(TorrentVersion.hybrid, id="hybrid"),
    ]
)
def torrent_version(request) -> TorrentVersion:
    return request.param


def _equal_sizes(path: Path, n: int, total_size: int) -> None:
    each_size = total_size // n
    for i in range(n):
        with open(path / f"{i}.bin", "wb") as f:
            f.write(randbytes(each_size))


@pytest.fixture(
    scope="module",
    params=[
        pytest.param((1, 16 * size.KiB), id="files-1-size-16KiB"),
        pytest.param((1, 1 * size.GiB), id="files-1-size-1GiB"),
        pytest.param((1000, 16 * size.KiB), id="files-1000-size-16KiB"),
        pytest.param((1000, 1 * size.GiB), id="files-1000-size-1GiB"),
    ],
)
def torrent_equal_sizes(request, tmp_path_factory) -> tuple[Path, int, int]:
    n, size = request.param

    pth = tmp_path_factory.mktemp(f"{n}_{size}_equal")
    _equal_sizes(pth, n, size)
    yield pth, n, size
    shutil.rmtree(str(pth))


@pytest.fixture(
    params=[
        pytest.param(1 * size.MiB, id="piece-size-1MiB"),
        pytest.param(32 * size.MiB, id="piece-size-32MiB"),
    ]
)
def piece_size(request) -> int:
    return request.param


def test_benchmark_decode(benchmark, torrent_file):
    benchmark(Torrent.read, torrent_file)


def test_benchmark_decode_torf(benchmark, torrent_file, monkeypatch):
    monkeypatch.setattr(TorfTorrent, "MAX_TORRENT_FILE_SIZE", 100 * (2**20))
    if "v2" in str(torrent_file):
        pytest.xfail()
    benchmark(TorfTorrent.read, torrent_file)


def test_create(torrent_version, torrent_equal_sizes, piece_size, benchmark):
    path, n_files, total_size = torrent_equal_sizes
    if n_files > 1000 and piece_size >= (2 * size.MiB) and torrent_version == TorrentVersion.hybrid:
        pytest.skip("takes too long")
    if n_files == 1 and total_size == 16 * size.KiB and piece_size > 1 * size.MiB:
        pytest.skip("pointless")

    def _hash_and_bencode() -> None:
        _ = TorrentCreate(
            path_root=path,
            piece_length=piece_size,
            trackers=["https://example.com/announce"],
        ).generate(version=torrent_version)

    benchmark(_hash_and_bencode)


def test_create_libtorrent(torrent_version, torrent_equal_sizes, piece_size, benchmark):
    path, n_files, total_size = torrent_equal_sizes
    if n_files > 1000 and piece_size >= (2 * size.MiB) and torrent_version == TorrentVersion.hybrid:
        pytest.skip("takes too long")
    if n_files == 1 and total_size == 16 * size.KiB and piece_size > 1 * size.MiB:
        pytest.skip("pointless")

    def _hash_and_bencode() -> None:
        generated = TorrentCreate(
            path_root=path,
            piece_length=piece_size,
            trackers=["https://example.com/announce"],
        ).generate_libtorrent(version=torrent_version)
        # torrent_models generation method also revalidates, so this is fair
        _ = Torrent.from_decoded(generated)

    benchmark(_hash_and_bencode)



"""
Types used only in v2 (and hybrid) torrents
"""

import multiprocessing as mp
from copy import deepcopy
from dataclasses import dataclass
from functools import cached_property
from math import ceil
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, NotRequired, TypeAlias, cast
from typing import Literal as L

from pydantic import AfterValidator, BaseModel, PlainSerializer
from pydantic_core.core_schema import SerializationInfo
from typing_extensions import TypeAliasType, TypedDict

from torrent_models.compat import get_size
from torrent_models.const import BLOCK_SIZE
from torrent_models.types.common import (
    AbsPath,
    RelPath,
    SHA256Hash,
    _divisible_by_16kib,
    _power_of_two,
)

if TYPE_CHECKING:
    pass

V2PieceLength = Annotated[int, AfterValidator(_divisible_by_16kib), AfterValidator(_power_of_two)]
"""
Per BEP 52: "must be a power of two and at least 16KiB"
"""


def _serialize_v2_hash(value: bytes, info: SerializationInfo) -> bytes | str | list[str]:
    if info.context and info.context.get("mode") == "print":
        ret: str = value.hex()

        if info.context.get("hash_truncate"):
            ret = ret[0:8]
        # split layers
        if len(ret) > 64:
            return [ret[i : i + 64] for i in range(0, len(ret), 64)]
        else:
            return ret

    return value


def _sort_keys(value: dict) -> dict:
    res = {}
    for k in sorted(value.keys()):
        if isinstance(value[k], dict):
            res[k] = _sort_keys(value[k])
        else:
            res[k] = value[k]
    return res


PieceLayerItem = Annotated[bytes, PlainSerializer(_serialize_v2_hash)]
PieceLayersType = dict[SHA256Hash, PieceLayerItem]
FileTreeItem = TypedDict(
    "FileTreeItem", {"length": int, "pieces root": NotRequired[PieceLayerItem]}
)
_FileTreeType = TypeAliasType(
    "_FileTreeType", 'dict[bytes, dict[L[""], FileTreeItem] | _FileTreeType]'
)
FileTreeType: TypeAlias = Annotated[_FileTreeType, AfterValidator(_sort_keys)]


class MerkleTree(BaseModel):
    """
    Representation and computation of v2 merkle trees

    A v2 merkle tree is a branching factor 2 tree where each of the leaf nodes is a 16KiB block.

    Two layers of the tree are embedded in a torrent file:

    - the ``piece layer``: the hashes from ``piece length/16KiB`` layers from the leaves.
      or, the layer where each hash corresponds to a chunk of the file ``piece length`` long.
    - the tree root.

    Padding is performed in two steps:

    - For files whose size is not a multiple of ``piece length``,
      pad the *leaf hashes* with zeros
      (the hashes, not the leaf data, i.e. 32 bytes not 16KiB of zeros)
      such that there are enough blocks to complete a piece
    - For files there the number of pieces does not create a balanced merkle tree,
      pad the *pieces hashes* with identical piece hashes each ``piece length`` long
      s.t. their leaf hashes are all zeros, as above.

    These are separated to avoid computing hashes of zero's unnecessarily.

    References:
        - https://www.bittorrent.org/beps/bep_0052_torrent_creator.py
    """

    path: RelPath
    """Path within torrent file"""
    piece_length: int
    """Piece length, in bytes"""
    piece_hashes: list[SHA256Hash] | None
    """
    hashes of each piece (the nth later of the merkle tree, determined by piece length).
    
    When a file is smaller than a single piece, set explicitly to ``None``.
    """
    root_hash: bytes
    """Root hash of the tree"""

    leaf_hashes: list[SHA256Hash] | None = None
    """SHA256 hashes of 16KiB leaf segments, if present."""

    @classmethod
    def from_path(
        cls,
        path: Path,
        piece_length: V2PieceLength,
        path_root: AbsPath | None = None,
        n_processes: int = mp.cpu_count(),
        progress: bool = False,
        **kwargs: Any,
    ) -> "MerkleTree":
        """
        Create a MerkleTree and return it with computed hashes

        Args:
            path (Path): Relative path to a file within a torrent directory. If absolute,
                must be beneath ``path_root``
            piece_length (V2PieceLength): Piece length used for piece hashes
            path_root (Path): Absolute path that should serve as the root of the torrent,
                the `path` must be a relative or absolute path within it.
            n_processes (int): Number of processes to use while hashing,
                default is n_cpus
            progress (bool): Display progress while hashing
            kwargs: Passed to :class:`.V2Hasher`

        """
        from torrent_models.hashing.v2 import V2Hasher

        if path_root is None:
            path_root = path.parent.resolve()
        else:
            assert path_root.is_absolute(), "Path root must be absolute"

        if path.is_absolute():
            path = path.relative_to(path_root)

        hasher = V2Hasher(
            paths=[path],
            piece_length=piece_length,
            path_root=path_root,
            n_processes=n_processes,
            progress=progress,
            **kwargs,
        )
        hashes = hasher.process()
        trees = hasher.finish_trees(hashes)
        assert len(trees) == 1, "Multiple trees returned from a single file constructor!"
        return trees[0]


class MerkleTreeShape(BaseModel):
    """
    Helper class to calculate values when constructing a merkle tree,
    without needing to have a merkle tree itself.

    Separated so that :class:`.MerkleTree` could just be a
    validated representation of the merkle tree
    rather than being the thing that hashes one,
    while also being able to validate the tree.
    """

    file_size: int
    """size of the file for which the merkle tree would be calculated, in bytes"""
    piece_length: V2PieceLength
    """piece length of the merkle tree"""

    @property
    def blocks_per_piece(self) -> int:
        return self.piece_length // BLOCK_SIZE

    @cached_property
    def n_blocks(self) -> int:
        """Number of total blocks in the file (excluding padding blocks)"""
        return ceil(self.file_size / BLOCK_SIZE)

    @cached_property
    def n_pieces(self) -> int:
        """Number of pieces in the file (or 0, if file is < piece_length)"""
        n_pieces = self.file_size / self.piece_length
        if n_pieces < 1:
            return 0
        else:
            return ceil(n_pieces)

    @cached_property
    def n_pad_blocks(self) -> int:
        """
        Number of blank blocks required for padding when hashing.

        Not strictly equivalent to the remainder to the nearest piece size,
        because we skip hashing all the zero blocks when we don't need to.
        (e.g. when to balance the tree we need to compute a ton of empty piece hashes)
        """
        if self.n_pieces <= 1:
            total_blocks = 1 << (self.n_blocks - 1).bit_length()
            return total_blocks - self.n_blocks
        elif (remainder := self.n_blocks % self.blocks_per_piece) == 0:
            return 0
        else:
            return self.blocks_per_piece - remainder

    @cached_property
    def n_pad_pieces(self) -> int:
        """Number of blank pieces required to balance merkle tree"""
        if self.n_pieces < 1:
            return 0
        return (1 << (self.n_pieces - 1).bit_length()) - self.n_pieces

    def validate_leaf_count(self, n_leaf_hashes: int) -> None:
        """Ensure that we have the right number of leaves for a merkle tree"""
        if self.n_pieces == 0:
            # ensure that n_blocks is a power of two
            n = n_leaf_hashes
            assert (n & (n - 1) == 0) and n != 0, (
                "For files smaller than one piece, "
                "must pad number of leaf blocks with zero blocks so n leaves is a power of two. "
                f"Got {n_leaf_hashes} leaf hashes with blocks_per_piece {self.blocks_per_piece}"
            )
        else:
            assert n_leaf_hashes % self.blocks_per_piece == 0, (
                f"leaf hashes must be a multiple of blocks per piece, pad with zeros. "
                f"Got {n_leaf_hashes} leaf hashes with blocks_per_piece {self.blocks_per_piece}"
            )


class FileTree(BaseModel):
    """
    A v2 torrent file tree is like

    - `folder/file1.png`
    - `file2.png`

    .. code-block:: python

        {
            "folder": {
                "file1.png": {
                    "": {
                        "length": 123,
                        "pieces root": b"<hash>",
                    }
                }
            },
            "file2.png": {
                "": {
                    "length": 123,
                    "pieces root": b"<hash>",
                }
            }
        }

    """

    tree: FileTreeType

    @classmethod
    def flatten_tree(cls, tree: FileTreeType) -> dict[str, FileTreeItem]:
        """
        Flatten a file tree, mapping each path to the item description
        """
        return _flatten_tree(tree)

    @classmethod
    def unflatten_tree(cls, tree: dict[str, FileTreeItem]) -> FileTreeType:
        """
        Turn a flattened file tree back into a nested file tree
        """
        return _unflatten_tree(tree)

    @cached_property
    def flat(self) -> dict[str, FileTreeItem]:
        """Flattened FileTree"""
        return self.flatten_tree(self.tree)

    @classmethod
    def from_flat(cls, tree: dict[str, FileTreeItem]) -> "FileTree":
        return cls(tree=cls.unflatten_tree(tree))

    @classmethod
    def from_trees(cls, trees: list[MerkleTree], base_path: Path) -> "FileTree":
        flat = {}
        for tree in trees:
            flat[tree.path.as_posix()] = FileTreeItem(
                **{"pieces root": tree.root_hash, "length": get_size(base_path / tree.path)}
            )
        return cls.from_flat(flat)


def _flatten_tree(val: dict, parts: list[str] | list[bytes] | None = None) -> dict:
    # NOT a general purpose dictionary walker.
    out: dict[bytes | str, dict] = {}
    if parts is None:
        # top-level, copy the input value
        val = deepcopy(val)
        parts = []

    for k, v in val.items():
        if isinstance(k, bytes):
            k = k.decode("utf-8")
        if k in (b"", ""):
            if isinstance(k, bytes):
                parts = cast(list[bytes], parts)
                out[b"/".join(parts)] = v
            elif isinstance(k, str):
                parts = cast(list[str], parts)
                out["/".join(parts)] = v
        else:
            out.update(_flatten_tree(v, parts + [k]))
    return out


def _unflatten_tree(val: dict) -> dict:
    out: dict[str | bytes, dict] = {}
    for k, v in val.items():
        is_bytes = isinstance(k, bytes)
        if is_bytes:
            k = k.decode("utf-8")
        parts = k.split(b"/") if is_bytes else k.split("/")
        parts = [p for p in parts if p not in (b"", "")]
        nested_subdict = out
        for part in parts:
            if part not in nested_subdict:
                nested_subdict[part] = {}
                nested_subdict = nested_subdict[part]
            else:
                nested_subdict = nested_subdict[part]
        if is_bytes:
            nested_subdict[b""] = v
        else:
            nested_subdict[""] = v
    return out


@dataclass
class PieceLayers:
    """
    Constructor for piece layers, along with the file tree, from a list of files

    Constructed together since file tree is basically a mapping of paths to root hashes -
    they are joint objects
    """

    piece_length: int
    """piece length (hash piece_length/16KiB blocks per piece hash)"""
    piece_layers: PieceLayersType
    """piece layers: mapping from root hash to concatenated piece hashes"""
    file_tree: FileTree

    @classmethod
    def from_trees(cls, trees: list[MerkleTree] | MerkleTree, base_path: Path) -> "PieceLayers":
        if not isinstance(trees, list):
            trees = [trees]
        lengths = [t.piece_length for t in trees]
        assert all(
            [lengths[0] == ln for ln in lengths]
        ), "Differing piece lengths in supplied merkle trees!"
        piece_length = lengths[0]
        piece_layers = {
            tree.root_hash: b"".join(tree.piece_hashes) for tree in trees if tree.piece_hashes
        }
        file_tree = FileTree.from_trees(trees, base_path)
        return PieceLayers(
            piece_length=piece_length, piece_layers=piece_layers, file_tree=file_tree
        )

    @classmethod
    def from_paths(
        cls,
        paths: list[RelPath],
        piece_length: int,
        path_root: Path,
        n_processes: int = mp.cpu_count(),
        progress: bool = False,
        **kwargs: Any,
    ) -> "PieceLayers":
        """
        Hash all the paths, construct the piece layers and file tree

        Args:
            paths (list[Path]): List of relative paths within some path root
            piece_length (V2PieceLength): piece length valid for v2 torrents
            path_root (Path): Root directory that contains ``paths``
            n_processes (int): number of processes to use for parallel processing.
                Default is n_cpus
            progress (bool): Display progress while hashing
            kwargs: passed to :class:`.V2Hasher`

        """
        from torrent_models.hashing.v2 import V2Hasher

        hasher = V2Hasher(
            paths=paths,
            path_root=path_root,
            piece_length=piece_length,
            n_processes=n_processes,
            progress=progress,
            **kwargs,
        )
        hashes = hasher.process()
        trees = hasher.finish_trees(hashes)
        return cls.from_trees(trees, base_path=path_root)



import os.path
import sys
from pathlib import Path


def get_size(path: Path) -> int:
    """
    Windows, helpfully, reports different sizes to stat,
    so we have to use the os.path.getsize() function instead.

    This *may* not be necessary, but is left in place in case we need
    compatibility for windows sizes in the future,
    which can be impacted by the NTFS filesystem.
    """
    if sys.platform == "win32":
        return os.path.getsize(path)
    else:
        return path.stat().st_size



from pathlib import Path

import libtorrent
import pytest

from torrent_models import Torrent
from torrent_models.types.common import TorrentVersion

from .conftest import DATA_DIR

ALL_TORRENTS = list(tf for tf in DATA_DIR.rglob("*.torrent") if "giant" not in tf.name)


def test_parse_hybrid():
    torrent = Torrent.read("tests/data/qbt_directory_hybrid.torrent")


@pytest.mark.parametrize(
    "tfile",
    [pytest.param(tf, id=str(tf.name)) for tf in ALL_TORRENTS],
)
def test_infohash(tfile: Path):
    """
    Test that the infohash that we get from a torrent is the same as what libtorrent would compute
    """
    lt_torrent = libtorrent.load_torrent_file(str(tfile))
    t = Torrent.read(tfile)

    if t.torrent_version in (TorrentVersion.v1, TorrentVersion.hybrid):
        assert t.v1_infohash is not None
        assert lt_torrent.info_hashes.has_v1()
        assert t.v1_infohash == lt_torrent.info_hashes.v1.to_bytes()
    if t.torrent_version in (TorrentVersion.v2, TorrentVersion.hybrid):
        assert t.v2_infohash is not None
        assert lt_torrent.info_hashes.has_v2()
        assert t.v2_infohash == lt_torrent.info_hashes.v2.to_bytes()
    if t.torrent_version not in TorrentVersion:
        raise ValueError("Torrent not detecting version")



import pytest

from torrent_models.types.v2 import FileTree

# add test cases of paired flattened and unflattened file trees here:

FILE_TREES = (
    {
        "directory": {
            "file1.exe": {"": {"length": 1, "pieces root": "sup"}},
            "subdir": {"file2.exe": {"": {"length": 2, "pieces root": "sup"}}},
        },
        "file3.exe": {"": {"length": 3, "pieces root": "sup"}},
    },
)

FLAT_FILE_TREES = (
    {
        "directory/file1.exe": {"length": 1, "pieces root": "sup"},
        "directory/subdir/file2.exe": {"length": 2, "pieces root": "sup"},
        "file3.exe": {"length": 3, "pieces root": "sup"},
    },
)


@pytest.mark.parametrize("tree,flat", zip(FILE_TREES, FLAT_FILE_TREES))
def test_flatten_tree(tree, flat):
    """
    We can flatten a tree!
    """
    assert FileTree.flatten_tree(tree) == flat


@pytest.mark.parametrize("tree,flat", zip(FILE_TREES, FLAT_FILE_TREES))
def test_unflatten_tree(tree, flat):
    """
    We can unflatten a tree!
    """
    assert FileTree.unflatten_tree(flat) == tree


@pytest.mark.parametrize("tree,flat", zip(FILE_TREES, FLAT_FILE_TREES))
def test_roundtrip_tree(tree, flat):
    """
    We can roundtrip flatten and unflattening trees!
    """
    assert FileTree.unflatten_tree(FileTree.flatten_tree(tree)) == tree
    assert FileTree.flatten_tree(FileTree.unflatten_tree(flat)) == flat



from typing import Literal, TypedDict


class ValidationContext(TypedDict, total=False):
    """
    Validation Context parameters that affect validation behavior

    Validation behavior can be controlled like this:

    ```python
    Torrent.model_validate(data, context={"padding": "default"})
    ```

    with any of the values described in this typed dict.

    This model is primarily used for documentation,
    because there is not a valid way that i am aware of to annotate the ValidationInfo.context
    dict since it is read-only (so cast can't be used),
    and one can't annotate non-self attributes.
    plz PR if you know how to get mypy to validate with this type.

    References:
        https://docs.pydantic.dev/latest/concepts/validators/#validation-context
    """

    padding: Literal["default", "ignore", "strict", "forbid"]
    """
    Control how padfiles are validated
    
    `'default`': v1-only torrents -> ignore
                 Hybrid torrents -> strict
    `'ignore'`: skip all padfile validation.
    `'strict'`: every file must start at a piece boundary.
    `'forbid'`: no padfiles may be present.

    if we are validating in pydantic's `strict` mode,
    - `'ignore'` is ignored: strict means that the torrent must be as correct as it can be
    - `'strict'` is unchanged, and becomes the default
    - `'forbid'` is unchanged for v1-only: it's strictly valid to forbid padfiles,
      but for hybrid torrents becomes an error: hybrid torrents *must* be padded
      and the padding must be correct.
    """
    padding_path: Literal["default", "strict"]
    """
    BEP0047 recommends that padfiles are named `.pad/{length}`:
    
    > While clients implementing this extensions will have no use for the path of a padding file 
    > it should be included for backwards compatibility since it is a mandatory field in BEP 3.
    > The recommended path is [".pad", "N"] where N is the length of the padding file in base10. 
    > This way clients not aware of this extension will write the padding files 
    > into a single directory, 
    > potentially re-using padding files from other torrents also stored in that directory.
    
    However this is not required, and is often violated. 
    BEP0047 also recommends now requiring `path` at all for padfiles:
    
    > To eventually allow the path field to be omitted clients implementing this BEP 
    > should not require it to be present on padding files.
    
    So the default behavior is to not check the padfile names,
    but passing `strict` here validates they are `[".pad", length]`
    
    References:
        https://www.bittorrent.org/beps/bep_0047.html
    """



"""
Hybrid v1/v2 torrent creation

This is not a straightforward combination of v1 and v2 hashing
since each version of torrent has different optimization requirements.

Since v1 is just a linear set of hashes, and the pieces are much larger units,
we can read a larger buffer and feed the whole thing into a hashing process at once.
v2 works on 16KiB chunks always, so the tradeoff of reading and processing time is a bit different.

Hybrid torrents require us to do both, as well as generate padfiles,
so we use routines from the v1 and v2 but build on top of them.
"""

from functools import cached_property
from itertools import count
from multiprocessing.pool import AsyncResult
from multiprocessing.pool import Pool as PoolType
from pathlib import Path
from typing import cast

from pydantic import PrivateAttr, field_validator

from torrent_models.const import BLOCK_SIZE
from torrent_models.hashing.base import Chunk, Hash
from torrent_models.hashing.v1 import V1Hasher
from torrent_models.hashing.v2 import V2Hasher, sort_v2
from torrent_models.types.v1 import FileItem
from torrent_models.types.v2 import PieceLayers, V2PieceLength


def add_padfiles(files: list[FileItem], piece_length: int) -> list[FileItem]:
    """
    Modify a v1 file list to intersperse .pad files
    """
    padded = []
    for f in files:
        padded.append(f)
        if f.attr in (b"p", "p"):
            continue
        if (remainder := f.length % piece_length) != 0:
            pad_length = piece_length - remainder
            pad = FileItem(length=pad_length, path=[".pad", str(pad_length)], attr=b"p")
            padded.append(pad)
    return padded


class HybridHasher(V1Hasher, V2Hasher):
    piece_length: V2PieceLength
    read_size: V2PieceLength | None = None

    _v1_chunks: list[Chunk] = PrivateAttr(default_factory=list)
    _last_path: Path | None = None
    _v1_counter: count = PrivateAttr(default_factory=count)

    @field_validator("paths", mode="after")
    def sort_paths(cls, value: list[Path]) -> list[Path]:
        """
        v1 torrents have arbitrary file sorting,
        but we mimick libtorrent/qbittorrent's sort order for consistency's sake
        """
        value = sort_v2(value)
        return value

    @cached_property
    def blocks_per_piece(self) -> int:
        return int(self.piece_length / BLOCK_SIZE)

    @cached_property
    def total_hashes(self) -> int:
        return self._v2_total_hashes() + self._v1_total_hashes_hybrid()

    def update(self, chunk: Chunk, pool: PoolType) -> list[AsyncResult]:
        res = self._update_v2(chunk, pool)
        res.extend(self._update_v1(chunk, pool))
        return res

    def _on_file_end(self, pool: PoolType) -> list[AsyncResult]:
        """Pad and submit buffer"""
        # breakpoint()
        if len(self._buffer) == 0:
            return []
        self._buffer.extend(bytes(self.piece_length - len(self._buffer)))
        self._last_path = cast(Path, self._last_path)
        chunk = Chunk.model_construct(
            idx=next(self._v1_counter), path=self._last_path, chunk=bytes(self._buffer)
        )
        self._buffer = bytearray()
        return [pool.apply_async(self._hash_v1, args=(chunk, self.path_root))]

    def _submit_v1(self, pool: PoolType) -> AsyncResult:
        piece = b"".join([c.chunk for c in self._v1_chunks])

        # append padding
        self._last_path = cast(Path, self._last_path)
        piece = b"".join([piece, bytes(self.piece_length - len(piece))])
        chunk = Chunk.model_construct(idx=next(self._v1_counter), path=self._last_path, chunk=piece)
        return pool.apply_async(self._hash_v1, args=(chunk, self.path_root))

    def _after_read(self, pool: PoolType) -> list[AsyncResult]:
        """Submit any remaining v1 pieces from the last file"""
        res = self._on_file_end(pool)
        return res

    def split_v1_v2(
        self,
        hashes: list[Hash],
    ) -> tuple[PieceLayers, list[bytes]]:
        """Split v1 and v2 hashes, returning sorted v1 pieces and v2 piece layers"""
        v1_pieces = [h for h in hashes if h.type == "v1_piece"]
        v1_pieces = sorted(v1_pieces, key=lambda h: h.idx)
        v1_pieces = [h.hash for h in v1_pieces]

        v2_leaf_hashes = [h for h in hashes if h.type == "block"]
        trees = self.finish_trees(v2_leaf_hashes)
        layers = PieceLayers.from_trees(trees, self.path_root)
        return layers, v1_pieces



import string
from itertools import product
from pathlib import Path
from random import randbytes, randint, random

from torrent_models import Torrent, TorrentCreate, TorrentVersion

try:
    import libtorrent
except ImportError:
    libtorrent = None  # type: ignore


def make_paths(n: int = 10, p_descend: float = 0.3, p_ascend: float = 0.1) -> list[Path]:
    if p_descend < p_ascend:
        raise ValueError("p_descend must be greater than p_ascend")

    paths = []
    letters = iter("".join([*letters]) for letters in product(string.ascii_letters, repeat=4))
    # ensure one root path
    paths.append(Path(next(letters)))
    if n == 1:
        return paths
    parent: Path | None = None
    while len(paths) < n:
        path = parent / next(letters) if parent else Path(next(letters))
        roll = random()
        if roll <= p_descend:
            parent = path
            continue
        elif roll <= p_ascend:
            parent = None
            continue
        paths.append(path)

    return paths


def make_files(
    base: Path, paths: list[Path] | None = None, piece_length: int = 32 * (2**10)
) -> list[Path]:
    if paths is None:
        paths = make_paths()

    paths = [base / path for path in paths]
    paths[0].parent.mkdir(exist_ok=True, parents=True)
    paths[0].write_bytes(randbytes(randint(0, piece_length - 1)))
    if len(paths) == 1:
        return paths
    paths[1].parent.mkdir(exist_ok=True, parents=True)
    paths[1].write_bytes(randbytes(randint(piece_length + 1, 3 * piece_length)))
    for path in paths[2:]:
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(randbytes(randint(0, 3 * piece_length)))
    return paths


def default_tcreate(
    paths: list[Path], base: Path, piece_length: int = 32 * (2**10)
) -> TorrentCreate:
    return TorrentCreate(
        piece_length=piece_length,
        paths=paths,
        path_root=base,
        trackers=["udp://example.com:6969/announce", "http://example.com/announce"],
        similar=[bytes(20)],
        webseeds=["https://example.com/data", "https://example.com/data2"],
    )


def default_torrent(
    paths: list[Path],
    base: Path,
    piece_length: int = 32 * (2**10),
    version: TorrentVersion = TorrentVersion.hybrid,
) -> Torrent:
    creator = default_tcreate(paths, base, piece_length)
    return creator.generate(version=version)


def default_libtorrent(
    paths: list[Path],
    base: Path,
    piece_length: int = 32 * (2**10),
    version: TorrentVersion = TorrentVersion.hybrid,
    output: Path | None = None,
    bencode: bool = False,
) -> dict | bytes:
    creator = default_tcreate(paths, base, piece_length)
    return creator.generate_libtorrent(version=version, output=output, bencode=bencode)



from torrent_models.const import BLOCK_SIZE, DEFAULT_TORRENT_CREATOR, GiB, KiB, MiB, TiB, size
from torrent_models.types import (
    FileItem,
    FileTree,
    FileTreeItem,
    FileTreeType,
    MerkleTree,
    MerkleTreeShape,
    PieceLayerItem,
    PieceLayers,
    PieceLayersType,
    Pieces,
    SHA1Hash,
    SHA256Hash,
    TorrentVersion,
    V1PieceLength,
    V2PieceLength,
)
from torrent_models.info import (
    InfoDictHybrid,
    InfoDictHybridCreate,
    InfoDictRoot,
    InfoDictV1,
    InfoDictV1Create,
    InfoDictV2,
    InfoDictV2Create,
)
from torrent_models.torrent import Torrent, pprint
from torrent_models.create import TorrentCreate

__all__ = [
    "BLOCK_SIZE",
    "DEFAULT_TORRENT_CREATOR",
    "GiB",
    "KiB",
    "MiB",
    "TiB",
    "FileItem",
    "FileTree",
    "FileTreeItem",
    "FileTreeType",
    "InfoDictHybrid",
    "InfoDictHybridCreate",
    "InfoDictRoot",
    "InfoDictV1",
    "InfoDictV1Create",
    "InfoDictV2",
    "InfoDictV2Create",
    "MerkleTree",
    "MerkleTreeShape",
    "PieceLayerItem",
    "PieceLayers",
    "PieceLayersType",
    "Pieces",
    "SHA1Hash",
    "SHA256Hash",
    "Torrent",
    "TorrentCreate",
    "TorrentVersion",
    "V1PieceLength",
    "V2PieceLength",
    "size",
    "pprint",
]



import hashlib
from collections import defaultdict
from functools import cached_property
from itertools import count
from multiprocessing.pool import AsyncResult
from multiprocessing.pool import Pool as PoolType
from pathlib import Path

from pydantic import PrivateAttr, field_validator

from torrent_models.compat import get_size
from torrent_models.const import BLOCK_SIZE
from torrent_models.hashing.base import Chunk, Hash, HasherBase
from torrent_models.types import SHA256Hash
from torrent_models.types.v2 import MerkleTree, MerkleTreeShape, V2PieceLength


class V2Hasher(HasherBase):
    piece_length: V2PieceLength
    read_size: V2PieceLength | None = None

    _v2_counter: count = PrivateAttr(default_factory=count)

    @classmethod
    @field_validator("read_size", mode="after")
    def read_size_is_block_size(cls, value: int) -> int:
        assert value == BLOCK_SIZE
        return value

    def _update_v2(self, chunk: Chunk, pool: PoolType) -> list[AsyncResult]:
        chunks = [
            Chunk.model_construct(
                path=chunk.path, chunk=chunk.chunk[i : i + BLOCK_SIZE], idx=next(self._v2_counter)
            )
            for i in range(0, len(chunk.chunk), BLOCK_SIZE)
        ]
        return [pool.apply_async(self._hash_v2, (c, self.path_root)) for c in chunks]

    def update(self, chunk: Chunk, pool: PoolType) -> list[AsyncResult]:
        return self._update_v2(chunk, pool)

    @cached_property
    def total_hashes(self) -> int:
        return self._v2_total_hashes()

    @classmethod
    def hash_root(
        cls,
        hashes: list[SHA256Hash],
    ) -> bytes:
        """
        Given hashes within a v2 merkle tree, compute their root.

        References:
            - https://www.bittorrent.org/beps/bep_0052_torrent_creator.py
        """
        assert len(hashes) & (len(hashes) - 1) == 0, "Must pass a balanced number of pieces"

        while len(hashes) > 1:
            hashes = [
                hashlib.sha256(left + right).digest() for left, right in zip(*[iter(hashes)] * 2)
            ]
        return hashes[0]

    def finish_trees(self, hashes: list["Hash"]) -> list["MerkleTree"]:
        """
        Create from a collection of leaf hashes.

        If leaf hashes from multiple paths are found, return a list of merkle trees.

        This method does *not* check that the trees are correct and complete -
        it assumes that the collection of leaf hashes passed to it is already complete.
        So e.g. it does not validate that the number of leaf hashes matches that which
        would be expected given the file size.

        Args:
            hashes (list[Hash]): collection of leaf hashes, from a single or multiple files
        """

        leaf_hashes = [h for h in hashes if h.type == "block"]
        leaf_hashes = sorted(leaf_hashes, key=lambda h: (h.path, h.idx))
        file_hashes = defaultdict(list)
        for h in leaf_hashes:
            file_hashes[h.path].append(h)

        trees = []
        for path, hashes in file_hashes.items():
            file_size = get_size(self.path_root / path)
            shape = MerkleTreeShape(file_size=file_size, piece_length=self.piece_length)
            hash_bytes = [h.hash for h in hashes]
            if len(hash_bytes) < shape.n_blocks + shape.n_pad_blocks:
                hash_bytes += [bytes(32)] * shape.n_pad_blocks

            piece_hashes = self.hash_pieces(hash_bytes, shape)
            if piece_hashes is None:
                root_hash = self.get_root_hash(hash_bytes, shape)
            else:
                root_hash = self.get_root_hash(piece_hashes, shape)

            tree = MerkleTree(
                path=path,
                piece_length=self.piece_length,
                leaf_hashes=hash_bytes,
                piece_hashes=piece_hashes,
                root_hash=root_hash,
            )
            trees.append(tree)
        return trees

    def hash_pieces(
        self, leaf_hashes: list[SHA256Hash], shape: MerkleTreeShape
    ) -> list[bytes] | None:
        """Compute the piece hashes for the layer dict"""
        if shape.n_pieces <= 1:
            return None

        shape.validate_leaf_count(len(leaf_hashes))
        # with concurrent.futures.ProcessPoolExecutor(max_workers=self.n_processes) as executor:
        piece_hashes = [
            self.hash_root(leaf_hashes[idx : idx + shape.blocks_per_piece])
            for idx in range(0, len(leaf_hashes), shape.blocks_per_piece)
        ]
        return piece_hashes

    def get_root_hash(self, piece_hashes: list[SHA256Hash], shape: MerkleTreeShape) -> bytes:
        """
        Compute the root hash, including any zero-padding pieces needed to balance the tree.

        If n_pieces == 0, the root hash is just the hash tree of the blocks,
        padded with all-zero blocks to have enough blocks for a full piece.

        So if `shape.n_pieces == 0`, then the hashes passed in should be the
        *leaf hashes* (since there are no piece hashes)
        """
        if shape.n_pieces <= 1:
            return self.hash_root(piece_hashes)

        if len(piece_hashes) == 1:
            return piece_hashes[0]

        if len(piece_hashes) == shape.n_pieces and shape.n_pad_pieces > 0:
            pad_piece_hash = self.hash_root([bytes(32)] * shape.blocks_per_piece)
            piece_hashes = piece_hashes + ([pad_piece_hash] * shape.n_pad_pieces)
        elif len(piece_hashes) != shape.n_pieces + shape.n_pad_pieces:
            raise ValueError(
                f"Expected either {shape.n_pieces} (unpadded) piece hashes or "
                f"{shape.n_pieces + shape.n_pad_pieces} hashes "
                f"(with padding for merkle tree balance). "
                f"Got: {len(piece_hashes)}"
            )

        return self.hash_root(piece_hashes)


def sort_v2(paths: list[Path]) -> list[Path]:
    """
    V2 paths are sorted in tree order, alphabetically.

    Mostly important for hybrid torrents, because v2 file trees are intrinsically sorted
    by the bencoding format
    """
    return sorted(paths, key=lambda f: f.as_posix())



from torrent_models.hashing.base import HasherBase
from torrent_models.hashing.hybrid import HybridHasher, add_padfiles
from torrent_models.hashing.v1 import V1Hasher
from torrent_models.hashing.v2 import V2Hasher

__all__ = ["HasherBase", "HybridHasher", "V1Hasher", "V2Hasher", "add_padfiles"]
